------------------------------------------------------------------------------
-- All code here contains stuff layout stuff, such as button positioning,	--
-- Button locking, scaling and so forth buttondragging is also handled here	--
------------------------------------------------------------------------------
local Layout = FlexBar2:NewModule("Layout");
-- If this var is incremented, SetDefaults() will be called for this module thus resetting all its settings
Layout.Version = 2;

-- Create a button mixin
Layout.ButtonMixin = FlexBar2.OO.Mixin
{
	"SetPosition",
	"Show",
	"Hide",
    "Lock",
    "Unlock",
    "SetIcon",
    "GetIcon",
    "SetScale",
    "GetScale",
    "StartDragging",
    "StopDragging",
};

-- This is called while loading the mixin
function Layout.ButtonMixin:Initialize()
	FlexBar2:Debug("               Setting up layout");

	-- Create the actual frame, inheriting from SecureActionButtonTemplate. Set a name for it, as sadly, some blizzard functions require a frame name instead of a reference
	local Frame = CreateFrame("CheckButton", "FB2_BTN_" .. self.uid, UIParent, "SecureActionButtonTemplate");
	self.Frame = Frame;
    Frame.Object = self;
	
	-- Strata and sizing, drag registering
	Frame:SetFrameStrata("MEDIUM");
    Frame:SetFrameLevel(1);
    Frame:SetWidth(36);
    Frame:SetHeight(36);

	-- Textures setup
	Frame:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress");
	Frame:SetCheckedTexture("Interface\\Buttons\\CheckButtonHilight");
	Frame:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square");
	Frame:GetHighlightTexture():SetBlendMode("ADD");
	Frame:GetCheckedTexture():SetBlendMode("ADD");

	-- Icon setup
	Frame.Icon = self.Frame:CreateTexture(nil, "BACKGROUND")
	Frame.Icon:SetAllPoints(Frame);

	-- Cooldown ticker setup
	Frame.Cooldown = CreateFrame("Cooldown", nil, self.Frame, "CooldownFrameTemplate");

	-- NormalTexture setup
	self.Frame:SetNormalTexture("Interface\\Buttons\\UI-Quickslot2");
	local NormalTexture = self.Frame:GetNormalTexture();
	NormalTexture:SetWidth(66);
    NormalTexture:SetHeight(66);
    NormalTexture:ClearAllPoints();
    NormalTexture:SetPoint("CENTER", 1, -1);

	-- Button settings
	Frame:RegisterForClicks("AnyUp");
	Frame:EnableMouse(true);
	Frame:SetClampedToScreen(true);
	Frame:SetMovable(true);

    -- Dragger frame
    Frame.Dragger = CreateFrame("Frame");
    local Dragger = Frame.Dragger;
    Dragger.Object = self;
    Dragger:SetFrameStrata("MEDIUM");
    Dragger:SetFrameLevel(3);
    Dragger:SetAllPoints(Frame);
    Dragger.texture = Dragger:CreateTexture(nil,'Overlay');
    Dragger.texture:SetTexture(0, 0.6, 0, 0.5);
    Dragger.texture:SetAllPoints(Dragger);
	Dragger:EnableMouse(true);
	Dragger:SetClampedToScreen(true);
	Dragger:SetMovable(false);

	-- Declare the fifth text field at CENTER
    local ButtonNameField = self.Frame.Dragger:CreateFontString(nil, "ARTWORK");
    ButtonNameField:SetFontObject("NumberFontNormal");
    ButtonNameField:SetJustifyH("CENTER");
    ButtonNameField:SetPoint("CENTER", self.Frame, "CENTER");
	ButtonNameField:SetText(self.Name);

	-- Set scripts for dragging
    Dragger:SetScript("OnMouseUp", function() self:StopDragging(); end);
	Dragger:SetScript("OnMouseDown", function()
		-- Start dragging
		self:StartDragging();
		-- If there's an action on the cursor, place it on the button
		self:PlaceAction();
	end);
	-- Support to receive actions while button is unlocked
	self.Frame.Dragger:SetScript("OnReceiveDrag", function() self:PlaceAction(); end);
end

-- Deactivator functions, in case the button gets deactivated, hide it
function Layout.ButtonMixin:Deactivate()
	self:Hide();
end

-- SetDefaults function, restore everything back to normal
function Layout.ButtonMixin:SetDefaults()
	self.InfoTable.Show = true;
    self.InfoTable.Lock = true;
    local ButtonNumber = tonumber(self.Name);
	-- Little conditional, if the BUtton's name matches a number, then set its position according to it, this makes everything look prettyer when using numbers as button name
    if(ButtonNumber) then
        self.InfoTable.X = 300 + (40 * (ButtonNumber - ((math.ceil(ButtonNumber / 12) - 1) * 12) - 1));
        self.InfoTable.Y = 300 + (40 * (math.ceil(ButtonNumber / 12) - 1));
    else
        self.InfoTable.X = 300;
    	self.InfoTable.Y = 300;
    end
	self.InfoTable.Scale = 1;
end

-- Apply the saved position, status & scale settings
function Layout.ButtonMixin:LoadSettings()
	self:SetPosition(self.InfoTable.X, self.InfoTable.Y);
	if(self.InfoTable.Show == true) then
		self:Show();
	else
		self:Hide();
	end
    self:SetScale(self.InfoTable.Scale or 1);
end

-- Start dragging method
function Layout.ButtonMixin:StartDragging()
	-- Check if the button has a group
    if(type(self.GroupMemberList) == "table") then
		-- Get this frame's positioning
        local _, _, _, MasterX, MasterY = self.Frame:GetPoint();
		-- Loop through all group members
        for _, GroupMember in pairs(self.GroupMemberList) do
			-- Only do something if the group member is not this button, this is important because we cant anchor a button to itself
            if(GroupMember ~= self) then
                local SlaveFrame = GroupMember.Frame;
				-- Get the group member's position
                local _, _, _, SlaveX, SlaveY = SlaveFrame:GetPoint();
				-- Set the member's position relative to this frame
                SlaveFrame:SetPoint("TOPLEFT", self.Frame, "TOPLEFT", SlaveX - MasterX, SlaveY - MasterY);
            end
        end
    end
	-- DRAG !
    self.Frame:StartMoving();
end

function Layout.ButtonMixin:StopDragging()
	-- Check if the button has a group
    if(type(self.GroupMemberList) == "table") then
        FlexBar2:Debug("Groupstop");
		-- Get this frame's positioning
        local _, _, _, MasterX, MasterY = self.Frame:GetPoint();
		-- Loop through all group members
        for _, GroupMember in pairs(self.GroupMemberList) do
			-- Only do something if the group member is not this button, this is important because we cant anchor a button to itself
            if(GroupMember ~= self) then
                FlexBar2:Debug("Stopping dragging for [", _ , "]");
                local SlaveFrame = GroupMember.Frame;
				-- Get the group member's positioning
                local _, _, _, SlaveX, SlaveY = SlaveFrame:GetPoint();
				-- Set position back relative to UIParent now we stopped dragging
                SlaveFrame.Object:SetPosition(SlaveX + MasterX, -(MasterY + SlaveY));
                SlaveFrame.Object.InfoTable.X = SlaveX + MasterX;
                SlaveFrame.Object.InfoTable.Y = -(MasterY + SlaveY);
            end
        end
    end
	-- STOP DRAGGING !
    self.Frame:StopMovingOrSizing();
	-- Lets save
    local _, _, _, X, Y = self.Frame:GetPoint();
    self.InfoTable.X = X;
    self.InfoTable.Y = -Y;
end

function Layout.ButtonMixin:SetPosition(X, Y)
	if(type(X) ~= "number") then X = self.InfoTable.X end
	if(type(Y) ~= "number") then Y = self.InfoTable.Y end
    self.Frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", X, -Y);
end

function Layout.ButtonMixin:Show()
	self.Frame:Show();
	-- Showstats has to be set to actualy show the button in all states, if we dont do this there is a chance it gets hidden on a state change
	self:SetAttribute("showstates", "*", false);
	-- Restore lock settings
    if(self.InfoTable.Lock == true) then
        self:Lock();
    else
        self:Unlock();
    end
end

function Layout.ButtonMixin:Hide()
	self.Frame:Hide();
	-- make sure the dragger frame is hidden
    self:Lock();
	-- Showstates has to be set to hide the button in all states, if we dont do this there is a chance it gets shown on a state change
	self:SetAttribute("showstates", "!*", false);
end

function Layout.ButtonMixin:Lock()
    self.Frame.Dragger:Hide();
end

function Layout.ButtonMixin:Unlock()
	if(self.Frame:IsShown()) then
		self.Frame.Dragger:Show();
	end
end

function Layout.ButtonMixin:GetIcon(Icon)
    return self.Frame.Icon:GetTexture(Icon);
end

function Layout.ButtonMixin:SetIcon(Icon)
    self.Frame.Icon:SetTexture(Icon);
end

function Layout.ButtonMixin:SetScale(Scale)
    self.Frame:SetScale(Scale);
end

function Layout.ButtonMixin:GetScale()
    self.Frame:GetScale();
end

-- GUI --
local Waterfall = FlexBar2.Waterfall;
local Positioning_X, Positioning_Y, Show, Status_Show, Status_Lock, Positioning_AutoAlign, TempScale;
function Layout.ButtonWaterfall(Action, ButtonName, ButtonTable)
	if(Action == "edit") then
		-- Retrieve those settings when the GUI gets shown
		Status_Show = FlexBar2:GetSetting(ButtonTable, "Show");
        Status_Lock = FlexBar2:GetSetting(ButtonTable, "Lock");
		-- Show button --
		Waterfall:AddControl(
		"type","checkbox",
		"text","Show button",
		"width",300,
		"causeRefresh", false,
		"getFunc" ,function() return Status_Show; end,
		"setFunc", function(Show) Status_Show = Show; end);
		-- Lock button --
        Waterfall:AddControl(
        "type","checkbox",
		"text","Lock button",
		"width",300,
		"causeRefresh", false,
		"getFunc" ,function() return Status_Lock; end,
		"setFunc", function(Lock) Status_Lock = Lock; end);
		-- Positioning textfield--
		Waterfall:AddControl(
		"type", "label",
		"text", "Button Position");
		Waterfall:AddControl(
		"type", "label",
		"text", "X: ");
		Waterfall:AddControl(
		"type", "textbox",
		"width", "50",
		"noNewLine", true,
		"causeRefresh", false,		
		"changedFunc", function(X) Positioning_X = X; end,
		"getFunc", function() if(#ButtonTable >= 2) then return "" else return FlexBar2:GetSetting(ButtonTable, "X") or 300; end end);
		Waterfall:AddControl(
		"type", "label",
		"text", "Y: ");
		Waterfall:AddControl(
		"type", "textbox",
		"width", "50",
		"noNewLine", true,
		"causeRefresh", false,		
		"changedFunc", function(Y) Positioning_Y = Y; end,
		"getFunc",  function() if(#ButtonTable >= 2) then return "" else return FlexBar2:GetSetting(ButtonTable, "Y") or 300; end end);
		-- Button scale text field --
        Waterfall:AddControl(
		"type", "label",
		"text", "Button Scale: ");
        Waterfall:AddControl(
		"type", "textbox",
		"width", "50",
		"noNewLine", true,
		"causeRefresh", false,		
		"changedFunc", function(Scale) TempScale = Scale; end,
		"getFunc", function() return FlexBar2:GetSetting(ButtonTable, "Scale") or 1 end);
		-- Special case, show an autoalign checkbox if we are configuring 2 buttons are more, autoalign causes all buttons to be ligned up next to eachother upon repositioning instead of being stacked on top of another
		if(#ButtonTable >= 2) then
			Waterfall:AddControl(
			"type","checkbox",
			"text","AutoAlign",
			"width",300,
			"causeRefresh", false,
			"getFunc" ,function() return Positioning_AutoAlign; end,
			"setFunc", function(AutoAlign) Positioning_AutoAlign = AutoAlign; end);
		end
	elseif(Action == "save") then
        local ButtonList = FlexBar2.Buttons;
		-- Make sure positioning settings have actualy been set before doing position stuff
		if(Positioning_X ~= "" and Positioning_Y ~= "") then
			-- Autoalign is true, realign the buttons based on the coords given
			if(Positioning_AutoAlign == true) then
                local X = Positioning_X + (40 * (ButtonName - ((math.ceil(ButtonName / 12) - 1) * 12) - 1));
                local Y = Positioning_Y + (40 * (math.ceil(ButtonName / 12) - 1));
                FlexBar2:ApplySetting(ButtonTable, "X", X);
                FlexBar2:ApplySetting(ButtonTable, "Y", Y);
                ButtonList[ButtonTable]:SetPosition(X, Y);
			-- Autoalign is false, just stack everything on top of another without caring how much buttons there are
			else
				FlexBar2:ApplySetting(ButtonTable, "X", Positioning_X);
				FlexBar2:ApplySetting(ButtonTable, "Y", Positioning_Y);
                ButtonList[ButtonTable]:SetPosition(Positioning_X, Positioning_Y);
			end 
		end
		-- Save the scale & apply it
		local Scale = tonumber(TempScale);
        FlexBar2:ApplySetting(ButtonTable, "Scale", Scale);
        ButtonList[ButtonTable]:SetScale(Scale);
		-- Save the status settings & apply em
        FlexBar2:ApplySetting(ButtonTable, "Show", Status_Show);
        if(Status_Show == true) then
            ButtonList[ButtonTable]:Show();
        else
            ButtonList[ButtonTable]:Hide();
        end
		-- Save lock settings & apply em
        FlexBar2:ApplySetting(ButtonTable, "Lock", Status_Lock);
        if(Status_Lock == true) then
            ButtonList[ButtonTable]:Lock();
        else
            ButtonList[ButtonTable]:Unlock();
        end
	end
end

function Layout.GeneralWaterfall()
	Waterfall:AddControl(
    "type","button",
    "text","Unlock all buttons",
    "fullRefresh", true,
    "width",150,
    "execFunc", function()
		for ButtonName, Button in pairs(FlexBar2.Buttons) do
			Button.InfoTable.Lock = false;
			Button:Unlock();
		end
    end);

	Waterfall:AddControl(
    "type","button",
    "text","Lock all buttons",
    "fullRefresh", true,
	"noNewLine", true,
    "width",120,
    "execFunc", function()
		for ButtonName, Button in pairs(FlexBar2.Buttons) do
			Button.InfoTable.Lock = true;
			Button:Lock();
		end
    end);
end

-- SlashCommands --
local SlashCommands = FlexBar2.SlashCommands;
-- /fb show command
local function SlashCommand_Show(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	FlexBar2:ApplySetting(ButtonTable, "Show", true);
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:Show();
	end
end

-- /fb hide command
local function SlashCommand_Hide(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	FlexBar2:ApplySetting(ButtonTable, "Show", false);
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:Hide();
	end
end

-- /fb lock command
local function SlashCommand_Lock(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	FlexBar2:ApplySetting(ButtonTable, "Lock", true);
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:Lock();
	end
end

-- /fb unlock command
local function SlashCommand_Unlock(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	FlexBar2:ApplySetting(ButtonTable, "Lock", false);
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:Unlock();
	end
end

-- /fb setposition command
local function SlashCommand_SetPosition(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	if(Params.smart == "false") then
		FlexBar2:ApplySetting(ButtonTable, "X", Params.x);
		FlexBar2:ApplySetting(ButtonTable, "Y", Params.y);
	elseif(Params.smart == "true") then
		for Number, Button in ipairs(ButtonTable) do
			FlexBar2:ApplySetting(Button, "X", Params.x + (40 * (Number - ((math.ceil(Number / 12) - 1) * 12) - 1)));
			FlexBar2:ApplySetting(Button, "Y", Params.y + (40 * (math.ceil(Number / 12) - 1)));
		end
	end
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:SetPosition();
	end
end

--  /fb setscale command
local function SlashCommand_SetScale(Params)
	local ButtonTable = FlexBar2:ToButtonTable(Params.button);
	local Scale = tonumber(Params.scale);
	FlexBar2:ApplySetting(ButtonTable, "Scale", Scale);
	local ButtonList = FlexBar2.Buttons;
	for _, ButtonName in ipairs(ButtonTable) do
		ButtonList[ButtonName]:SetScale(tonumber(Scale));
	end
end

-- Localize this neat verify function for faster access
local VerifyExistance = FlexBar2.SlashCommands.VerifyExistance;
-- AddCommand format :AddCommand(Command, Function, Group, Help)
-- AddParam format :AddParam(Command, Param, Required, AllowedValues, DefaultValue, VerifyFunc) 
-- everything here should be self explainative knowing the syntax
SlashCommands:AddCommand("show", SlashCommand_Show, "layout", "/FlexBar2 show button=<Button>");
	SlashCommands:AddParam("show", "button", true, nil, nil, VerifyExistance)
SlashCommands:AddCommand("hide", SlashCommand_Hide, "layout", "/FlexBar2 hide button=<Button>");
	SlashCommands:AddParam("hide", "button", true, nil, nil, VerifyExistance)
SlashCommands:AddCommand("lock", SlashCommand_Lock, "layout", "/FlexBar2 lock button=<Button>");
	SlashCommands:AddParam("lock", "button", true, nil, nil, VerifyExistance)
SlashCommands:AddCommand("unlock", SlashCommand_Unlock, "layout", "/FlexBar2 unlock button=<Button>");
	SlashCommands:AddParam("unlock", "button", true, nil, nil, VerifyExistance)
SlashCommands:AddCommand("setposition", SlashCommand_SetPosition, "layout", "/FlexBar2 setposition button=<Button> x=<x coordinate> y=<y coordinate> (smart=<true/false>)");
	SlashCommands:AddParam("setposition", "button", true, nil, nil, VerifyExistance)
	SlashCommands:AddParam("setposition", "x", false, nil, "300", nil);
	SlashCommands:AddParam("setposition", "y", false, nil, "300", nil);
	SlashCommands:AddParam("setposition", "smart", false, {"true", "false"}, "true", nil);
SlashCommands:AddCommand("setscale", SlashCommand_SetScale, "layout", "/FlexBar2 setscale button=<Button> scale=<Scale>");
    SlashCommands:AddParam("setscale", "button", true, nil, nil, VerifyExistance)
    SlashCommands:AddParam("setscale", "scale", false, nil, "1", function(Scale) if(string.find(Scale, "^%d*%.?%d+$")) then return true; else return "Invalid scale"; end end);
